/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "headless_vga.hpp"
#include "vga_regmap.h"

using namespace std;
using namespace sc_core;

namespace {
  const sc_time period(1000000 / VGA_FREQUENCY, SC_US);
  const sc_time period_third(1000000 / (3*VGA_FREQUENCY), SC_US);

  const double TRANSFERING = 0.6;
  const double WAITING = 0.25;

  const sc_time time_between_word = sc_time(1, SC_US);
}

HeadlessVGA::HeadlessVGA(sc_core::sc_module_name):
  PwtModule(this),
  m_address(0), m_intr(false), m_waiting(false)
{
  INFO(2, "in constructor");

  initiator.set_dmi(&m_dmi);

  SC_THREAD(thread);

  irq.initialize(false);
}

HeadlessVGA::~HeadlessVGA() {}

void HeadlessVGA::thread() {
  while (true) {
    if (m_address != 0) {
      vsync();
    } else {
      wait(address_event);
    }
  }
}

void HeadlessVGA::vsync() {
  INFO(2, "transfer and draw");
  set_activity(TRANSFERING);
  draw();
  wait(period-period_third);
  set_activity(WAITING);

  INFO(2, "vsync, raise IRQ");
  m_intr = true;
  irq.write(true); trace_irq();
  m_waiting = true;
  wait(period_third);
  irq.write(false);
  m_waiting = false;
}

void HeadlessVGA::draw() {
  const uint32_t last_address =
    m_address + ((VGA_WIDTH + (VGA_HEIGHT * VGA_LINE)) / 8) - 4;
  bool has_dmi = m_dmi.has_dmi(m_address, last_address);
  if (!has_dmi) {
    m_dmi.get_dmi(m_address, *initiator.get_base_port().get_interface(0));
    has_dmi = m_dmi.has_dmi(m_address, last_address);
  }
  sc_time local_offset = SC_ZERO_TIME;
  if (has_dmi) {
    INFO(2, "Read image using DMI");
    for (int y = 0; y < VGA_HEIGHT; ++y) {
      for (int x = 0; x < VGA_WIDTH; x += 32) {
        const uint32_t address
          = m_address + ((x + (y * VGA_LINE)) / 8);
        m_dmi.apply_side_effects(1, tlm::TLM_READ_COMMAND, local_offset);
        gui_draw(x, y, m_dmi[address]);
        local_offset += time_between_word;
      }
    }
  } else {
    INFO(2, "Read image using transactions");
    for (int y = 0; y < VGA_HEIGHT; ++y) {
      for (int x = 0; x < VGA_WIDTH; x += 32) {
        const uint32_t address
          = m_address + ((x + (y * VGA_LINE)) / 8);
        uint32_t data;
        initiator.read(address, data, ~0, local_offset);
        gui_draw(x, y, data);
        local_offset += time_between_word;
      }
    }
  }
  assert(local_offset<=period-period_third && "failed to read image before deadline");
}

tlm::tlm_response_status
HeadlessVGA::read(uint32_t addr, uint32_t& data)
{
  switch (addr) {
  case VGA_CFG_OFFSET:
    data = m_address;
    break;
  case VGA_STT_OFFSET:
    data = 0xFFFFFFFF;
    break;
  case VGA_INT_OFFSET:
    data = m_intr;
    break;
  case VGA_WAITING_OFFSET:
    data = m_waiting;
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

tlm::tlm_response_status
HeadlessVGA::write(uint32_t addr, uint32_t data)
{
  switch (addr) {
  case VGA_CFG_OFFSET:
    m_address = data;
    if (m_address!=0)
      address_event.notify(SC_ZERO_TIME); // FIXME: align on period
    INFO(2, "VGA_CFG_OFFSET changed to " << hex << m_address << dec);
    break;
  case VGA_STT_OFFSET:
    /* see VHDL */
    break;
  case VGA_INT_OFFSET:
    m_intr = false;
    break;
  default:
    WARNING("register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

void HeadlessVGA::gui_draw(int, int, uint32_t) {
  // nothing in base class
}
