/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "memory.hpp"
#include "trace.hpp"
#include "untimed_ext.hpp"

using namespace std;
using namespace sc_core;

namespace {
  ParameterBool dmi_disabled("no-dmi",
                             "Disable Direct Memory Interface");
}

const unsigned Memory::READ_LATENCY;
const unsigned Memory::WRITE_LATENCY;

// Constructor
Memory::Memory(sc_module_name, unsigned int byte_size):
  PwtTargetModule(this, .5, .8, .5, .8),
  m_size(byte_size),
  m_storage(new uint32_t[byte_size/4]),
  m_debug(false),
  m_dmi_allowed(!dmi_disabled.get())
{
  INFO(2, "in constructor");

  target.register_b_transport(this, &Memory::b_transport);
  target.register_transport_dbg(this, &Memory::transport_dbg);
  target.register_get_direct_mem_ptr(this,
                                     &Memory::get_direct_mem_ptr);
}

// Destructor
Memory::~Memory() {
  INFO(2, "in destructor");
  delete[] m_storage;
}

void Memory::b_transport(tlm::tlm_generic_payload &payload,
                         sc_core::sc_time &delay_time)
{
  tlm::tlm_command cmd = payload.get_command();
  if (cmd==tlm::TLM_IGNORE_COMMAND) {
    payload.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
    return;
  }
  if (payload.get_streaming_width()!=4) {
    WARNING("streaming width not supported: "
            << payload.get_streaming_width());
    payload.set_response_status(tlm::TLM_BURST_ERROR_RESPONSE);
    return;
  }
  const uint64_t address = payload.get_address();
  const unsigned length = payload.get_data_length();
  if (address+length>m_size) {
    WARNING("access outside the memory range");
    payload.set_response_status(tlm::TLM_ADDRESS_ERROR_RESPONSE);
    return;
  }
  if (address%4!=0) {
    WARNING("non-aligned addresses are not supported");
    payload.set_response_status(tlm::TLM_ADDRESS_ERROR_RESPONSE);
    return;
  }
  uint32_t *data =
    reinterpret_cast<uint32_t*>(payload.get_data_ptr());
  uint32_t *byte_enable =
    reinterpret_cast<uint32_t*>(payload.get_byte_enable_ptr());
  uint32_t all_one = ~0;
  unsigned byte_enable_length = 4;
  if (byte_enable==NULL)
    byte_enable = &all_one;
  else
    byte_enable_length = payload.get_byte_enable_length();
  assert(byte_enable_length>0 && "illegal byte enable length");
  if (byte_enable_length%4!=0) {
    WARNING("non-aligned byte enable length are not supported");
    payload.set_response_status(tlm::TLM_BYTE_ENABLE_ERROR_RESPONSE);
    return;
  }
  for (unsigned i = 0, bi = 0, ei = length/4; i!=ei; ++i, ++bi) {
    if (bi==byte_enable_length/4) bi = 0;
    b_transport_word(cmd, address+i*4, data[i], byte_enable[bi]);
  }
  if (!m_debug) {
    untimed_extension *ext;
    payload.get_extension(ext);
    if (ext) {
      add_delayed_side_effects(ext->get_size(), cmd, delay_time);
    } else {   
      apply_side_effects(length/4, cmd, delay_time);
      delay_time +=
        (cmd==tlm::TLM_READ_COMMAND ? READ_LATENCY : WRITE_LATENCY) * get_period();
    }
  }
  payload.set_response_status(tlm::TLM_OK_RESPONSE);
  payload.set_dmi_allowed(m_dmi_allowed);
}

void Memory::b_transport_word(const tlm::tlm_command command,
                              const uint32_t address,
                              uint32_t &data,
                              const uint32_t be_mask)
{
  uint32_t &local_data = m_storage[address/4];
  if (command==tlm::TLM_WRITE_COMMAND) {
    local_data &=~ be_mask;
    local_data |= data&be_mask;
  } else {
    data &=~ be_mask;
    data |= local_data&be_mask;
  }
}

unsigned int Memory::transport_dbg(tlm::tlm_generic_payload &payload)
{
  sc_time zero = SC_ZERO_TIME;
  m_debug = true;
  this->b_transport(payload, zero);
  m_debug = false;
  if (payload.get_response_status()==tlm::TLM_OK_RESPONSE)
    return payload.get_data_length();
  else
    return 0;
}

bool Memory::get_direct_mem_ptr(tlm::tlm_generic_payload &payload,
                                tlm::tlm_dmi&  dmi_data)
{
  const uint64_t address = payload.get_address();
  if (address<m_size && m_dmi_allowed) {
    dmi_data.allow_read_write();
    dmi_data.set_start_address(0);
    dmi_data.set_end_address(m_size-1);
    dmi_data.set_dmi_ptr(reinterpret_cast<unsigned char*>(m_storage));
    dmi_data.set_read_latency(READ_LATENCY*get_period());
    dmi_data.set_write_latency(WRITE_LATENCY*get_period());
    side_effect_extension *ext;
    payload.get_extension(ext);
    if (ext) ext->add_side_effect(this);
    return true;
  } else {
    dmi_data.set_start_address(address);
    dmi_data.set_end_address(address);
    return false;
  }
}

void Memory::set_frequency_hook(unsigned) {
  target->invalidate_direct_mem_ptr(0, m_size-1);
}
