/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "timer.hpp"
#include "timer_regmap.h"
#include "bit.hpp"
#include "trace.hpp"

using namespace std;
using namespace sc_core;

Timer::Timer(sc_module_name, sc_time period):
  PwtModule(this),
  m_period(period)
{
  SC_METHOD(timer0);
  dont_initialize();
  sensitive << m_csr_event[0];

  SC_METHOD(timer1);
  dont_initialize();
  sensitive << m_csr_event[1];

  SC_THREAD(interrupt);
  for (int timer = 0; timer < 2; ++timer) {
    m_csr[timer] = 0;
    m_tlr[timer] = 0;
    m_tcr[timer] = 0;
  }
  irq.initialize(false);
}

void Timer::interrupt() {
  while (true) {
    wait(m_irq_event);
    INFO(1, "interrupt");
    irq.write(true); trace_irq();
    wait(m_period);
    irq.write(false);
  }
}

void Timer::timer(int timer) {
  if (m_refresh[timer]) {
    m_refresh[timer] = false;
  refresh:
    if (TEST_BIT(m_csr[timer], TIMER_UDT)) { // down
      m_csr_event[timer].notify
        (m_period * (m_tcr[timer] + 2));
    } else { // up
      m_csr_event[timer].notify
        (m_period * (~0 - m_tcr[timer] + 2));
    }
    return;
  }
  SET_BIT(m_csr[timer], TIMER_TINT); // interrupt
  if (TEST_BIT(m_csr[timer], TIMER_ENIT)) { // enable interrupt
    m_irq_event.notify();
  }
  if (TEST_BIT(m_csr[timer], TIMER_ARHT)) { // auto reload
    m_tcr[timer] = m_tlr[timer];
    goto refresh;
  } else { // hold
    if (TEST_BIT(m_csr[timer], TIMER_UDT)) { // down
      m_tcr[timer] = -1;
    } else { // up
      m_tcr[timer] = 0;
    }
  }
}

void Timer::timer0() {
  timer(0);
}

void Timer::timer1() {
  timer(1);
}

tlm::tlm_response_status
Timer::read(uint32_t addr, uint32_t& data)
{
  int timer = 0;
  switch (addr) {
  case TIMER_0_CSR_OFFSET:
  CSR:
    data = m_csr[timer];
    break;
  case TIMER_0_TLR_OFFSET:
  TLR:
    data = m_tlr[timer];
    break;
  case TIMER_0_TCR_OFFSET:
  TCR:
    data = m_tcr[timer];
    break;
  case TIMER_1_CSR_OFFSET:
    timer = 1; goto CSR;
  case TIMER_1_TLR_OFFSET:
    timer = 1; goto TLR;
  case TIMER_1_TCR_OFFSET:
    timer = 1; goto TCR;
  default:
    SC_REPORT_ERROR(name(), "register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}

tlm::tlm_response_status
Timer::write(uint32_t addr, uint32_t data)
{
  int timer = 0;
  switch (addr) {
  case TIMER_0_CSR_OFFSET: 
  CSR:
    if (TEST_BIT(data, TIMER_TINT)) { // Interrupt
      CLEAR_BIT(data, TIMER_TINT);
      CLEAR_BIT(m_csr[timer], TIMER_TINT);
      INFO(2, "clean interrupt");
    }
    if (TEST_BIT(data, TIMER_ENT)) { // Enable Timer
      CLEAR_BIT(data, TIMER_ENT);
      SET_BIT(m_csr[timer], TIMER_ENT);
      m_refresh[timer] = true;
      m_csr_event[timer].notify();
      INFO(2, "enable timer #" << timer);
    } else { // Disable Timer
      if(TEST_BIT(m_csr[timer], TIMER_ENT)) { // was enabled
        CLEAR_BIT(m_csr[timer], TIMER_ENT);
        m_csr_event[timer].cancel();
        INFO(2, "disable timer #" << timer);
      }
    }
    if (TEST_BIT(data, TIMER_ENIT)) { // Enable Interrupt
      CLEAR_BIT(data, TIMER_ENIT);
      SET_BIT(m_csr[timer], TIMER_ENIT);
    } else { // Disable Interrupt
      CLEAR_BIT(m_csr[timer], TIMER_ENIT);
    }
    if (TEST_BIT(data, TIMER_LOAD)) { // Load
      CLEAR_BIT(data, TIMER_LOAD);
      if (TEST_BIT(m_csr[timer], TIMER_ENT)) { // enabled
        m_refresh[timer] = true;
        m_csr_event[timer].notify();
      }
      INFO(2, "load timer #" << timer);
      m_tcr[timer] = m_tlr[timer];
    }
    if (TEST_BIT(data, TIMER_ARHT)) { // Auto Reload
      CLEAR_BIT(data, TIMER_ARHT);
      SET_BIT(m_csr[timer], TIMER_ARHT);
    } else { // Hold Timer
      CLEAR_BIT(m_csr[timer], TIMER_ARHT);
    }
    if (TEST_BIT(data, TIMER_UDT)) { // Up
      CLEAR_BIT(data, TIMER_UDT);
      SET_BIT(m_csr[timer], TIMER_UDT);
    } else { // Down
      CLEAR_BIT(m_csr[timer], TIMER_UDT);
    }
    if (data) {
      SC_REPORT_WARNING(name(), "invalid bits in write to CSR");
    }
    break;
  case TIMER_0_TLR_OFFSET:
  TLR:
    m_tlr[timer] = data;
    break;
  case TIMER_0_TCR_OFFSET:
  TCR:
    SC_REPORT_ERROR(name(), "TCR is read only");
    break;
  case TIMER_1_CSR_OFFSET:
    timer = 1; goto CSR;
  case TIMER_1_TLR_OFFSET:
    timer = 1; goto TLR;
  case TIMER_1_TCR_OFFSET:
    timer = 1; goto TCR;
  default:
    SC_REPORT_ERROR(name(), "register not implemented");
    return tlm::TLM_ADDRESS_ERROR_RESPONSE;
  }
  return tlm::TLM_OK_RESPONSE;
}
