/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

#include "basic.hpp"
#include "mb_wrapper.hpp"
#include "microblaze.hpp"
#include "trace.hpp"
#include "parameter.hpp"
#include <iomanip>

using namespace std;
using namespace sc_core;

namespace {
  ParameterInt
  mb_quantum("mb-quantum",
             "MicroBlaze quantum for temporal decoupling (in ns)",
             1000);

  const double IDLE = 0.04;
  const double RUNNING = 0.6;
}

MBWrapper::MBWrapper(sc_module_name, unsigned id):
  PwtModule(this),
  irq("irq"),
  m_iss(id),
  m_id(id),
  m_clock_gated(false),
  m_irq_duration(-1),
  m_step_count(0),
  m_local_offset(SC_ZERO_TIME)
{
  INFO(1, "in constructor");
  
  initiator.set_dmi(&m_dmi);

  m_iss.reset();
  m_iss.setIrq(false);

  SC_THREAD(run_iss);

  SC_METHOD(process_irq);
  dont_initialize();
  sensitive << irq;

  debug_target.register_b_transport(this,
                                    &MBWrapper::b_transport);
  debug_target.register_transport_dbg(this,
                                      &MBWrapper::transport_dbg);
  debug_target.register_get_direct_mem_ptr(this,
                                     &MBWrapper::get_direct_mem_ptr);
}

MBWrapper::~MBWrapper() {
  INFO(1, "in destructor; step count = " << m_step_count);
}

void MBWrapper::process_irq(void) {
  INFO(2, "IRQ signal changed to " << irq.read());
  if (irq.read()) {
    m_iss.setIrq(true); trace_irq();
    m_irq_duration = 10;
    interrupt_event.notify();
  }
}

uint32_t MBWrapper::read_word_or_abort(uint32_t addr,
                                       const char * msg)
{
  if (m_dmi.has_dmi(addr)) {
    m_dmi.apply_side_effects(1, tlm::TLM_READ_COMMAND, m_local_offset);
    return m_dmi[addr];
  } else {
    return read_word_or_abort_noinline(addr, msg);
  }
}

uint32_t MBWrapper::read_word_or_abort_noinline(uint32_t addr, const char * msg) {
  uint32_t data;
  tlm::tlm_response_status status;
  status = initiator.read(addr, data, ~0, m_local_offset);
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
  if (initiator.is_dmi_allowed())
    m_dmi.get_dmi(addr, *initiator.get_base_port().get_interface(0));
  return data;
}

uint32_t MBWrapper::read_half_or_abort(uint32_t addr,
                                       const char * msg)
{
  uint32_t data;
  tlm::tlm_response_status status;
  const unsigned shift = 8*(2-(addr&2));
  const uint32_t be_mask = 0xffff<<shift;
  status = initiator.read(addr&~2, data, be_mask, m_local_offset);
  data >>= shift;
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
  return data&0xffff;
}

uint32_t MBWrapper::read_byte_or_abort(uint32_t addr,
                                       const char * msg)
{
  uint32_t data;
  tlm::tlm_response_status status;
  const unsigned shift = 8*(3-(addr&3));
  const uint32_t be_mask = 0xff<<shift;
  status = initiator.read(addr&~3, data, be_mask, m_local_offset);
  data >>= shift;
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
  return data;
}

void MBWrapper::write_word_or_abort(uint32_t addr,
                                    uint32_t data,
                                    const char * msg)
{
  if (m_dmi.has_dmi(addr)) {
    m_dmi.apply_side_effects(1, tlm::TLM_WRITE_COMMAND, m_local_offset);
    m_dmi[addr] = data;
  } else {
    write_word_or_abort_noinline(addr, data, msg);
  }
}

void MBWrapper::write_word_or_abort_noinline(uint32_t addr,
                                             uint32_t data,
                                             const char * msg)
{
  tlm::tlm_response_status status;
  status = initiator.write(addr, data, ~0, m_local_offset);
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
  if (initiator.is_dmi_allowed())
    m_dmi.get_dmi(addr, *initiator.get_base_port().get_interface(0));
}

void MBWrapper::write_half_or_abort(uint32_t addr,
                                    uint32_t data,
                                    const char * msg)
{
  tlm::tlm_response_status status;
  const unsigned shift = 8*(2-(addr&2));
  const uint32_t be_mask = 0xffff<<shift;
  status =
    initiator.write(addr&~2, data<<shift, be_mask, m_local_offset);
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
}

void MBWrapper::write_byte_or_abort(uint32_t addr,
                                    uint32_t data,
                                    const char * msg)
{
  tlm::tlm_response_status status;
  const unsigned shift = 8*(3-(addr&3));
  const uint32_t be_mask = 0xff<<shift;
  status =
    initiator.write(addr&~3, data<<shift, be_mask, m_local_offset);
  if (status!=tlm::TLM_OK_RESPONSE) {
    ERROR(msg);
  }
}

void MBWrapper::run_iss(void) {
  sc_core::sc_time quantum(sc_time(mb_quantum.get(), SC_NS));
  set_activity(RUNNING);
  while (true) {
    while (m_clock_gated) {
      INFO(2, "clock-gated, waiting for non-null frequency");
      wait(m_freq_event);
    }
    if (m_iss.m_sleeping) {
      if (m_local_offset>=SC_ZERO_TIME) {
        wait(m_local_offset);
        m_local_offset = SC_ZERO_TIME;
      }
      if (m_iss.m_sleeping) {
        // wait for the interrupt
        set_activity(IDLE);
        INFO(2, "idle, waiting for interrupt");
        wait(irq.posedge_event());
        set_activity(RUNNING);
        m_iss.m_sleeping = false;
      }
    }
    
    if (m_iss.isBusy()) {
      m_iss.nullStep();
    } else {
      bool instr_asked;
      uint32_t instr_addr;
      m_iss.getInstructionRequest(instr_asked, instr_addr);

      if (instr_asked) {
        /* The ISS requested an instruction.
         * We have to do the instruction fetch
         * by reading from memory. */
        
        INFO(3, "Fetch at address: " << hex << instr_addr <<dec); 
        const uint32_t localbuf =
          read_word_or_abort(instr_addr, "prefecth failed!");
        INFO(3, "Fetch: " << hex << localbuf <<dec); 
        m_iss.setInstruction(false, localbuf);
      }

      bool mem_asked;
      enum iss_t::DataAccessType mem_type;
      uint32_t mem_addr;
      uint32_t mem_wdata;
      m_iss.getDataRequest(mem_asked, mem_type, mem_addr, mem_wdata);

      if (mem_asked) {
        switch (mem_type) {
        case iss_t::READ_WORD: {
          // The ISS requested a data read (mem_addr into localbuf).
          const uint32_t localbuf =
            read_word_or_abort(mem_addr);
          m_iss.setDataResponse(false, localbuf);
        } break;

        case iss_t::READ_BYTE: {
          const uint32_t localbuf =
            read_byte_or_abort(mem_addr);
          m_iss.setDataResponse(0, localbuf);
        } break;

        case iss_t::READ_HALF: {
          const uint32_t localbuf =
            read_half_or_abort(mem_addr);
          m_iss.setDataResponse(0, localbuf);
        } break;
          
        case iss_t::WRITE_HALF: {
          write_half_or_abort(mem_addr, mem_wdata);
          m_iss.setDataResponse(false, 0);
        } break;

        case iss_t::WRITE_BYTE: {
          write_byte_or_abort(mem_addr, mem_wdata);
          m_iss.setDataResponse(false, 0);
        } break;

        case iss_t::WRITE_WORD: {
          write_word_or_abort(mem_addr, mem_wdata);
          m_iss.setDataResponse(false, 0);
        } break;

        case iss_t::LINE_INVAL:
        case iss_t::STORE_COND:
        case iss_t::READ_LINKED:
          std::cerr << name() << ": ERROR: Operation " << mem_type
                    << " unsupported.\n";
          abort();
          break;
        }
      }

      bool fsl_req; 
      enum iss_t::FSL_Access fsl_type;
      uint32_t fsl_port;
      uint32_t dataw; 
      m_iss.getFSLRequest(fsl_req, fsl_type, fsl_port, dataw);
      if (fsl_req) 
        switch (fsl_type) {
        case iss_t::BLOCK_GET_DATA:
          INFO(2, "Handling FSL blocking get operation");
          // FSL0 is reserved to proc_id 
          if (fsl_port == 0) {
            INFO(2, "Procid asked, return " << m_id);
            m_iss.setFSLResponse(0, m_id);
          } else {
            ERROR("FSL port not implemented"); abort();
          }
          break; 

        case iss_t::BLOCK_PUT_DATA:
          INFO(2, "Handling FSL blocking put operation");
          ERROR("FSL port not implemented"); abort();
          break;
        default: 
          ERROR("unsupported FSL access"); abort();
          break; 
        }

      m_iss.step();
      ++m_step_count;
      m_local_offset += get_period();
      if (m_local_offset>=quantum) {
        wait(m_local_offset);
        m_local_offset = SC_ZERO_TIME;
      }

      if (m_irq_duration > 0) {
        m_irq_duration--;
      } else if (m_irq_duration == 0) {
        m_iss.setIrq(false);
        m_irq_duration = -1;
      }
    }
  }
}

unsigned int
MBWrapper::transport_dbg(tlm::tlm_generic_payload& payload) {
  if (payload.get_streaming_width()==4) {
    if (payload.get_data_length()%4!=0) {
      payload.set_response_status(tlm::TLM_BURST_ERROR_RESPONSE);
      return 0;
    }
    return initiator->transport_dbg(payload);
  } else if (payload.get_streaming_width()==1) {
    const unsigned size = payload.get_data_length();
    const uint8_t *data = payload.get_data_ptr();
    const uint64_t address = payload.get_address();
    tlm::tlm_generic_payload trans;
    trans.set_streaming_width(4);
    trans.set_command(tlm::TLM_WRITE_COMMAND);
    uint32_t word;
    trans.set_data_ptr(reinterpret_cast<uint8_t*>(&word));
    trans.set_data_length(4);
    uint32_t be_mask;
    trans.set_byte_enable_ptr(reinterpret_cast<uint8_t*>(&be_mask));
    trans.set_byte_enable_length(4);
    unsigned sent = 0;
    for (unsigned i = 0; i<size; ++i) {
      const uint64_t addr = (address+i)&~3llu;
      trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      trans.set_address(addr);
      const unsigned shift = 8*(3-((address+i)&3));
      be_mask = 0xff<<shift;
      word = data[i]<<shift;
      const unsigned n = initiator->transport_dbg(trans);
      sent += n==4 ? 1 : 0;
      if (trans.get_response_status()!=tlm::TLM_OK_RESPONSE) {
        payload.set_response_status(trans.get_response_status());
        return sent;
      }
    }
    payload.set_response_status(tlm::TLM_OK_RESPONSE);
    return sent;
  } else {
    TODO("non-word accesses");
  }
}

void MBWrapper::b_transport(tlm::tlm_generic_payload &payload, sc_time&) {
  payload.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
  return;
}

bool MBWrapper::get_direct_mem_ptr(tlm::tlm_generic_payload &payload,
                        tlm::tlm_dmi &dmi_data)
{
  dmi_data.set_start_address(payload.get_address());
  dmi_data.set_end_address(payload.get_address());
  payload.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
  return false;
}

tlm::tlm_target_socket<32, tlm::tlm_base_protocol_types>&
MBWrapper::get_debug_target()
{
  return debug_target;
}

void MBWrapper::set_endianness(endian_t endian) {
  if (endian!=ProcessorIF::BIG_ENDIAN_32) {
    USER_ERROR("Endianness configuration not supported");
  }
}

void MBWrapper::set_pc(uint32_t new_pc) {
  m_iss.setDebugRegisterValue(32, new_pc);
  INFO(2, "set PC to 0x" << hex << new_pc <<dec);
}

void MBWrapper::set_frequency_hook(unsigned freq) {
  if (freq>0) {
    m_freq_event.notify();
    m_clock_gated = false;
  } else {
    m_clock_gated = true;
  }
}

// void MBWrapper::transition_hook(state_t from, state_t to) {
//   (void) from;
//   if (to.millivolt * FREQUENCE_MAIN < MILLIVOLT_MAIN * to.freq) {
//     ERROR("cannot support frequency " << to.freq/1000 << "kHz with voltage "
//           << to.millivolt << "mV");
//   }
// }
