/*
 * Copyright (C) 2009, 2010, 2011, 2012, 2013  by CNRS and Grenoble-INP
 *
 * This file is part of LIBTLMPWT
 *
 * LIBTLMPWT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Matthieu Moy <matthieu.moy@imag.fr>, Grenoble-INP
 *          Tayeb Bouhadiba, CNRS
 *          Claude Helmstetter, CNRS
 *
 * This software has been partially supported by the French ANR project HELP
 *                                                         (ANR-09-SEGI-006)
 */

// Code partly coming from SimSoC (copyright INRIA, L-GPL)

#ifndef _HOST_MEMORY_HPP_
#define _HOST_MEMORY_HPP_

#include "ecounters_ext.hpp"
#include "pwt_module.hpp"

class HostMemory {
  
  class DMI_Descriptor {
  public:
    DMI_Descriptor();

    DMI_Descriptor(uint32_t *memory,
                   uint32_t size,
                   uint32_t offset,
                   sc_core::sc_time read_latency,
                   sc_core::sc_time write_latency,
                   const std::vector<PwtTargetModule*> &side_effect);

    inline bool is_in_range(uint32_t address) const {
      return m_offset<=address && address<m_end;
    }
    inline bool is_in_range(uint32_t address,
                            uint32_t last_addr) const {
      return m_offset<=address && last_addr<m_end;
    }

  protected:
    friend class HostMemory;

    uint32_t *m_memory;
    uint32_t  m_size;
    uint32_t  m_offset;
    uint32_t  m_end;
    sc_core::sc_time m_read_latency;
    sc_core::sc_time m_write_latency;
    PwtTargetModule **m_side_effects; // NULL terminated
  }; // class DMI_Descriptor

public:
  HostMemory();

  void get_dmi(uint32_t addr,
               tlm::tlm_fw_direct_mem_if<tlm::tlm_generic_payload> &port);

  inline bool has_dmi(uint32_t address);
  inline bool has_dmi(uint32_t address, uint32_t last_addr);

  inline uint32_t& operator[](uint32_t addr) /*-*/ {
    return m_memory[(addr-m_offset)/4];
  }
  inline uint32_t  operator[](uint32_t addr) const {
    return m_memory[(addr-m_offset)/4];
  }

  inline void apply_side_effects(unsigned size, tlm::tlm_command cmd,
                                 sc_core::sc_time &delay);

  void clear();

protected:
  std::vector<DMI_Descriptor> dmi_descriptors;
  uint32_t *m_memory;
  uint32_t  m_offset;
  uint32_t  m_end;
  sc_core::sc_time m_read_latency;
  sc_core::sc_time m_write_latency;
  PwtTargetModule **m_side_effects; // NULL terminated
};

bool HostMemory::has_dmi(uint32_t address) {
  if (m_end>address && address>=m_offset)
    return true;
  for (size_t i = 0, ei = dmi_descriptors.size(); i!=ei; ++i) {
    if (dmi_descriptors[i].is_in_range(address)) {
      m_memory = dmi_descriptors[i].m_memory;
      m_offset = dmi_descriptors[i].m_offset;
      m_end = dmi_descriptors[i].m_end;
      m_read_latency = dmi_descriptors[i].m_read_latency;
      m_write_latency = dmi_descriptors[i].m_write_latency;
      m_side_effects = dmi_descriptors[i].m_side_effects;
      return true;
    }
  }
  return false;
}

bool HostMemory::has_dmi(uint32_t address, uint32_t last_addr) {
  if (m_end>last_addr && address>=m_offset)
    return true;
  for (size_t i = 0, ei = dmi_descriptors.size(); i!=ei; ++i) {
    if (dmi_descriptors[i].is_in_range(address, last_addr)) {
      m_memory = dmi_descriptors[i].m_memory;
      m_offset = dmi_descriptors[i].m_offset;
      m_end = dmi_descriptors[i].m_end;
      m_read_latency = dmi_descriptors[i].m_read_latency;
      m_write_latency = dmi_descriptors[i].m_write_latency;
      m_side_effects = dmi_descriptors[i].m_side_effects;
      return true;
    }
  }
  return false;
}

void HostMemory::apply_side_effects(unsigned size, tlm::tlm_command cmd,
                                    sc_core::sc_time &delay)
{
  if (pwt_time_stamp+delay<pwt_step_end) {
    if (cmd==tlm::TLM_READ_COMMAND) {
      for (PwtTargetModule ** i = m_side_effects; *i!=NULL; ++i)
        (*i)->apply_side_effects_at_0_read(size);
      delay += m_read_latency;
    } else {
      for (PwtTargetModule ** i = m_side_effects; *i!=NULL; ++i)
        (*i)->apply_side_effects_at_0_write(size);
      delay += m_write_latency;
    }
  } else {
    for (PwtTargetModule ** i = m_side_effects; *i!=NULL; ++i)
      (*i)->apply_side_effects(size, cmd, pwt_time_stamp+delay);
    delay += (cmd==tlm::TLM_READ_COMMAND) ? m_read_latency : m_write_latency;
  }
}

#endif // _HOST_MEMORY_HPP_
